program getnumthreads

  implicit none
  
  INTEGER :: nthreads,idx
  
  nthreads=-1
  
  CALL get_num_threads(nthreads)
  
  PRINT *, "----------"
  PRINT *, "Serial region  ",nthreads
  PRINT *
  !$OMP PARALLEL PRIVATE(nthreads,idx)
  
  CALL get_num_threads(nthreads)
  CALL get_thread_id(idx)

  !$OMP BARRIER 
  !$OMP SINGLE
  PRINT *, "Outer parallel ",nthreads
  PRINT *
  !$OMP END SINGLE

  !$OMP PARALLEL NUM_THREADS(idx+1) PRIVATE(nthreads)
  CALL get_num_threads(nthreads)
  !$OMP SINGLE
  PRINT *, "Inner parallel ",nthreads
  !$OMP END SINGLE
  !$OMP END PARALLEL
  
  !$OMP BARRIER
  !$OMP SINGLE
  PRINT *, "----------"
  !$OMP END SINGLE
  
  !$OMP END PARALLEL
end program getnumthreads

SUBROUTINE get_num_threads(nthreads)
  IMPLICIT NONE 

  INTEGER, INTENT(OUT) :: nthreads
  INTEGER, SAVE        :: accumulation = 0
  INTEGER, SAVE        :: spin         = 0
  
  nthreads=-1
  
  !!
  !! Only master enter, and keep spinning until other regions are finished
  !!
  !$OMP MASTER
  !$OMP CRITICAL
  DO WHILE(spin .ne. 0)
     
  END DO
  !$OMP ATOMIC
  spin = spin + 1
  !$OMP END CRITICAL
  !$OMP END MASTER

  !!
  !! All the slaves wait at this barrier for their master
  !!
  !$OMP BARRIER
  
  !!
  !! If a master pass all its team goes through the following
  !!
  !$OMP ATOMIC
  accumulation = accumulation+1  
  !$OMP BARRIER
  nthreads = accumulation  
  !$OMP BARRIER
  accumulation = 0
  !$OMP BARRIER

  !!
  !! The master of the current team can release the lock
  !!
  !$OMP MASTER
  !$OMP ATOMIC
  spin = spin - 1  
  !$OMP END MASTER
  !$OMP BARRIER

END SUBROUTINE get_num_threads

SUBROUTINE get_thread_id(idx)

  IMPLICIT NONE
  
  INTEGER, INTENT(OUT) :: idx
  INTEGER              :: ii,nthreads

  CALL get_num_threads(nthreads)
  !$OMP DO SCHEDULE(static)
  DO ii = 1,nthreads
     idx = ii -1
  END DO
  !$OMP END DO  

END SUBROUTINE get_thread_id
